/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_MPC_DEBUG			FALSE
#define FE_MPC_GRIP_DEBUG		FALSE

#include "operate/operate.pmh"

using namespace fe;

void ManipulatorCommon::initialize(void)
{
}

void ManipulatorCommon::bindOverlay(sp<DrawI>& a_spDrawOverlay)
{
	m_spOverlayCache=a_spDrawOverlay;
}

void ManipulatorCommon::handle(Record& a_rSignal)
{
#if FE_MPC_DEBUG
	feLog("ManipulatorCommon::handle\n");
#endif

	//* TODO fix step-behind draw access
	if(!m_spDrawCache.isValid())
	{
		return;
	}

	sp<ViewI> spView=m_spDrawCache->view();

	m_event.bind(windowEvent(a_rSignal));

	const Vector2i mouse(m_event.mouseX(),m_event.mouseY());

	if(!m_dragging)
	{
		if(m_hotGrip>=0 &&
				m_event.isMousePress(WindowEvent::e_itemLeft))
		{
			m_dragging=TRUE;
			m_lastGrip=m_hotGrip;
		}
		else if(m_lastGrip>=0 &&
				m_event.isMousePress(WindowEvent::e_itemMiddle))
		{
			m_dragging=TRUE;
			m_hotGrip=m_lastGrip;
		}
	}

	if(m_dragging)
	{
		FEASSERT(m_hotGrip>=0);
		if(m_event.isMouseDrag())
		{
			U32 item=m_event.item();
			BWORD shifted=FALSE;
			if(item&WindowEvent::e_keyShift)
			{
				shifted=TRUE;
				item^=WindowEvent::e_keyShift;
			}

			const Real gain=shifted? 0.1: 1.0;

			Grip& rGrip=m_gripArray[m_hotGrip];

			SpatialVector facing=rGrip.m_planeFacing;
			if(magnitudeSquared(facing)<0.5)
			{
//				feLog("\n grip %d change facing %s\n",
//						m_hotGrip,c_print(facing));

				Box2i viewport=spView->viewMode()->viewport();
				Vector2i center(viewport.size()[0]/2,viewport.size()[1]/2);

				facing=unitSafe(
						spView->unproject(center[0],center[1],1.0,
						ViewMode::e_perspective)-
						spView->unproject(center[0],center[1],0.0,
						ViewMode::e_perspective));
			}

			const SpatialVector lastOrigin=spView->unproject(
					m_lastMouse[0],m_lastMouse[1],0.0,
					ViewMode::e_perspective);
			const SpatialVector lastRay=unitSafe(spView->unproject(
					m_lastMouse[0],m_lastMouse[1],1.0,
					ViewMode::e_perspective)-lastOrigin);

			const SpatialVector currentOrigin=spView->unproject(
					mouse[0],mouse[1],0.0,ViewMode::e_perspective);
			const SpatialVector currentRay=unitSafe(spView->unproject(
					mouse[0],mouse[1],1.0,ViewMode::e_perspective)-
					currentOrigin);

			const Real lastDist=RayPlaneIntersect<Real>::solve(
					rGrip.m_planeCenter,facing,0.0,
					lastOrigin,lastRay);
			const Real currentDist=RayPlaneIntersect<Real>::solve(
					rGrip.m_planeCenter,facing,0.0,
					currentOrigin,currentRay);

			const SpatialVector lastHit=lastOrigin+lastDist*lastRay;
			const SpatialVector currentHit=currentOrigin+currentDist*currentRay;

			const SpatialVector movement=currentHit-lastHit;

#if FE_MPC_GRIP_DEBUG
			feLog("\n mouse %d %d\n",mouse[0],mouse[1]);
			feLog(" plane %s facing %s\n",
					c_print(rGrip.m_planeCenter),c_print(facing));
			feLog(" last %s  %s dist %.6G\n",
					c_print(lastOrigin),c_print(lastRay),lastDist);
			feLog(" current %s  %s dist %.6G\n",
					c_print(currentOrigin),c_print(currentRay),currentDist);
			feLog(" movement %s\n",c_print(movement));
			feLog(" lastHit %s\n",c_print(lastHit));
			feLog(" currentHit %s\n",c_print(currentHit));
#endif

			for(U32 pass=0;pass<2;pass++)
			{
				const Real scale=magnitude(rGrip.m_alignment[pass]);
				if(scale>0.0)
				{
					const Real along=dot(movement,rGrip.m_alignment[pass]);

#if FE_MPC_GRIP_DEBUG
					feLog("  pass %d along %.6G\n",pass,along);
					feLog("  align %s unitdot %.6G\n",
							c_print(rGrip.m_alignment[pass]),
							dot(unitSafe(movement),rGrip.m_alignment[pass]));
#endif

					const String& rKey=rGrip.m_key[pass];
					if(!rKey.empty())
					{
						catalog<Real>(rKey)+=along*gain;
					}
				}
			}
		}
		else
		{
			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(!buttons)
			{
				m_dragging=FALSE;
			}
		}
	}
	else
	{
		I32 bestIndex= -1;
		Real bestDist=0.0;

		const U32 gripCount=m_gripArray.size();
		for(U32 gripIndex=0;gripIndex<gripCount;gripIndex++)
		{
			Grip& rGrip=m_gripArray[gripIndex];

//			feLog("grip %d/%d %s radius %.6G align %s %s center %s facing %s\n",
//					gripIndex,gripCount,
//					c_print(rGrip.m_hotSpot),rGrip.m_radius,
//					c_print(rGrip.m_alignment[0]),
//					c_print(rGrip.m_alignment[1]),
//					c_print(rGrip.m_planeCenter),
//					c_print(rGrip.m_planeFacing));

			const Real gripDist=
					magnitudeSquared(SpatialVector(mouse)-rGrip.m_hotSpot);
			if(!gripIndex || gripDist<bestDist)
			{
				bestIndex=gripIndex;
				bestDist=gripDist;
			}
		}

//		feLog("best %d %.6G\n",bestIndex,bestDist);
		m_hotGrip=bestIndex;
	}

//	feLog("ManipulatorCommon::handle hot %d\n",m_hotGrip);

	m_lastMouse=mouse;
}

void ManipulatorCommon::draw(sp<DrawI> a_spDrawI,const Color *color)
{
#if FE_MPC_DEBUG
	feLog("ManipulatorCommon::draw\n");
#endif

	m_spDrawCache=a_spDrawI;
	updateGrips();

	if(m_hotGrip>=0)
	{
		const BWORD centered=TRUE;

		sp<ViewI> spView=m_spOverlayCache->view();
		sp<ViewMode> spViewMode=spView->viewMode();
		const Box2i viewport=spViewMode->viewport();

		const Color black(0.0,0.0,0.0,1.0);
		const Color blue(0.5,0.5,1.0,1.0);
		SpatialVector textPoint(0.5*viewport.size()[0],
				viewport.size()[1]-48,1.0);
		drawLabel(m_spOverlayCache,textPoint,
				m_gripArray[m_hotGrip].m_message,
				centered,3,blue,NULL,&black);

		textPoint=m_gripArray[m_hotGrip].m_hotSpot;
		drawLabel(m_spOverlayCache,textPoint,
				m_gripArray[m_hotGrip].m_tag,
				centered,3,blue,NULL,&black);
	}
}
