/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAB_DEBUG			FALSE

#include "surface.pmh"

namespace fe
{

Protectable* SurfaceAccessibleBase::clone(Protectable* pInstance)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::clone\n");
#endif

	SurfaceAccessibleBase* pSurfaceAccessibleBase= pInstance?
			dynamic_cast<SurfaceAccessibleBase*>(pInstance):
			new SurfaceAccessibleBase();

	//* HACK save and reload to duplicate RecordGroup

	FEASSERT(m_spScope.isValid());
	const String tempname="temp.rg";
	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,tempname);

	sp<RecordGroup>& rspRecordGroup=pSurfaceAccessibleBase->m_spRecordGroup;
	rspRecordGroup=RecordView::loadRecordGroup(m_spScope,tempname);

	//* update surface with new Record
	RecordGroup::iterator it=rspRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	Record record=spRA->getRecord(0);
	pSurfaceAccessibleBase->bind(record);

	return pSurfaceAccessibleBase;
}

BWORD SurfaceAccessibleBase::load(String a_filename)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::load\n");
#endif

	FEASSERT(m_spScope.isValid());

	m_spRecordGroup=RecordView::loadRecordGroup(m_spScope,a_filename);
	if(m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	RecordGroup::iterator it=m_spRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(spRA);

	Record record=spRA->getRecord(0);

	Instance model;
	model.set<Record>(registry()->master()->typeMaster(),record);
	bind(model);

	return TRUE;
}

BWORD SurfaceAccessibleBase::save(String a_filename)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::save\n");
#endif

	FEASSERT(m_spScope.isValid());
	FEASSERT(m_spRecordGroup.isValid());

	if(m_spScope.isNull() || m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,a_filename);
	return TRUE;
}

sp<SurfaceAccessorI> SurfaceAccessibleBase::accessor(
	Element a_element,String a_name,Creation a_create)
{
	sp<SurfaceAccessorBase> spAccessor(new SurfaceAccessorBase);

	if(spAccessor.isValid())
	{
		spAccessor->bind(m_record);
		if(spAccessor->bind(a_element,a_name))
		{
			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			return spAccessor;
		}
	}

	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleBase::accessor(
	Element a_element,Attribute a_attribute,Creation a_create)
{
	sp<SurfaceAccessorBase> spAccessor(new SurfaceAccessorBase);

	if(spAccessor.isValid())
	{
		spAccessor->bind(m_record);
		if(spAccessor->bind(a_element,a_attribute))
		{
			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			return spAccessor;
		}
	}

	return sp<SurfaceAccessorI>(NULL);
}

BWORD SurfaceAccessibleBase::discard(SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute)
{
	String name;
	switch(a_attribute)
	{
		case SurfaceAccessibleI::e_position:
			name="P";
			break;
		case SurfaceAccessibleI::e_normal:
			name="N";
			break;
		default:
			return FALSE;
	}

	return discard(a_element,name);
}

sp<SurfaceI> SurfaceAccessibleBase::surface(String a_group)
{
//	feLog("SurfaceAccessibleBase::surface(%s)\n",a_group.c_str());

	//* TODO select form
	//* TODO group
	sp<SurfaceTrianglesAccessible> spSurface=
			registry()->create("*.SurfaceTrianglesAccessible");
	spSurface->bind(sp<SurfaceAccessibleI>(this));
	return spSurface;
}

sp<SurfaceI> SurfaceAccessibleBase::subSurface(U32 a_subIndex,String a_group)
{
	//* TODO

	return sp<SurfaceI>(NULL);
}

sp<SurfaceRange> SurfaceAccessibleBase::atomize(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomize\n");
#endif

	if(a_atomicChange==e_pointsOfPrimitives)
	{
		return atomizeConnectivity(a_atomicChange,a_group,a_desiredCount);
	}

	//* fallback to non-atomic

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);
	const Element element=asPoints? e_point: e_primitive;
	sp<SurfaceAccessorI> spAccessor=accessor(element,e_generic);
	const I32 elementCount=spAccessor.isValid()? spAccessor->count(): 0;
	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	//* use everything

	if(a_group.empty())
	{
		if(elementCount>0)
		{
			spSurfaceRange->nonAtomic().add(0,elementCount-1);
		}
		return spSurfaceRange;
	}

	//* use group

	I32 spanStart= -1;
	SurfaceRange::MultiSpan& rMultiSpan=spSurfaceRange->nonAtomic();

	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const I32 elementPage=
				asPoints? pointPage(elementIndex): primitivePage(elementIndex);
		if(spanStart<0 && elementPage>=0)
		{
			//* begin span
			spanStart=elementIndex;
		}
		else if(spanStart>=0 && elementPage<0)
		{
			//* complete span
			rMultiSpan.add(spanStart,elementIndex-1);
			spanStart=-1;
		}
	}

	//* complete last span
	if(spanStart>=0)
	{
		rMultiSpan.add(spanStart,elementCount-1);

	}

	return spSurfaceRange;
}

sp<SurfaceRange> SurfaceAccessibleBase::atomizeConnectivity(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
	preparePaging(a_atomicChange,a_group);

	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	sp<SurfaceAccessorI> spPoints=accessor(e_point,e_position);
	sp<SurfaceAccessorI> spVertices=accessor(e_primitive,e_vertices);

	const U32 pointCount=spPoints->count();
	const U32 primitiveCount=spVertices->count();

	std::vector<U32> pointPatch(pointCount,0);
	std::vector<I32> primitivePatch(primitiveCount,0);

	std::vector<U32> pagePatch(primitiveCount+1,0);	//* pages <= prims+1
	std::vector<U32> histogram(primitiveCount+1,0);	//* patches <= prims+1

	U32 maxPatch=0;
	U32 maxPage=0;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		if(primitivePage(primitiveIndex)<0)
		{
			FEASSERT(primitiveIndex<primitiveCount);
			primitivePatch[primitiveIndex]= -1;
			continue;
		}

		U32 lowPatch=0;
		U32 lowPage=0;
		BWORD multiPatch=FALSE;
		BWORD zeroPatches=FALSE;
		BWORD multiPage=FALSE;

		const U32 subCount=spVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=spVertices->integer(primitiveIndex,subIndex);
			FEASSERT(pointIndex<pointCount);
			U32& rPatch=pointPatch[pointIndex];
			FEASSERT(rPatch<=maxPatch);

			const I32 page=pointPage(pointIndex);
			if(page>0)
			{
				FEASSERT(page<=primitiveCount);
				const U32 existingPatch=pagePatch[page];
				FEASSERT(existingPatch<=maxPatch);

				if(existingPatch)
				{
					rPatch=existingPatch;
				}
				if(!lowPage)
				{
					lowPage=page;
				}
				else if(lowPage!=page)
				{
					multiPage=TRUE;
					if(lowPage>page)
					{
						lowPage=page;
					}
				}
			}
			if(rPatch)
			{
				if(!lowPatch)
				{
					lowPatch=rPatch;
				}
				else if(lowPatch!=rPatch)
				{
					multiPatch=TRUE;
					if(lowPatch>rPatch)
					{
						lowPatch=rPatch;
					}
				}
			}
			else
			{
				zeroPatches=TRUE;
			}
		}

		if(multiPage)
		{
			//* leave for postAtomic
			lowPatch=0;
		}
		else if(multiPatch)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				const U32 patch=pointPatch[pointIndex];
				if(patch!=lowPatch)
				{
					for(U32 pointOther=0;pointOther<pointCount;pointOther++)
					{
						FEASSERT(pointOther<pointCount);
						U32& rOtherPatch=pointPatch[pointOther];

						if(rOtherPatch==patch)
						{
							//* merge all 'patch' into 'lowPatch'
							rOtherPatch=lowPatch;
						}
						else if(rOtherPatch==maxPatch)
						{
							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPatch=patch;
						}
					}
					FEASSERT(maxPatch);
					maxPatch--;
				}
			}
		}
		else if(!lowPatch)
		{
			maxPatch++;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=maxPatch;
			}
			lowPatch=maxPatch;
		}
		else if(zeroPatches)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=lowPatch;
			}
		}

		FEASSERT(lowPatch<=maxPatch);

		FEASSERT(primitiveIndex<primitiveCount);
		primitivePatch[primitiveIndex]=lowPatch;

//		feLog("primitive %d/%d lowPatch %d/%d\n",
//				primitiveIndex,primitiveCount,lowPatch,maxPatch);

		FEASSERT(lowPatch<=primitiveCount);
		histogram[lowPatch]+=subCount;

		if(lowPatch)
		{
			FEASSERT(lowPage<=primitiveCount);
			pagePatch[lowPage]=lowPatch;
		}
		if(maxPage<lowPage)
		{
			maxPage=lowPage;
		}
	}

#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPage %d\n",maxPage);
	for(U32 pageIndex=0;pageIndex<=maxPage;pageIndex++)
	{
		FEASSERT(pageIndex<=primitiveCount);
		feLog("page %d uses patch %d\n",pageIndex,pagePatch[pageIndex]);
	}
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPatch %d\n",maxPatch);
	for(U32 patchIndex=0;patchIndex<=maxPatch;patchIndex++)
	{
		FEASSERT(patchIndex<=primitiveCount);
		feLog("patch %d has %d points\n",patchIndex,histogram[patchIndex]);
	}
#endif

	const U32 desiredSize=a_desiredCount? primitiveCount/a_desiredCount: 1;

	std::vector<U32> patchAtom(maxPatch+1,0);

	I32 maxAtom=1;
	U32 atomSize=0;
	for(U32 patchIndex=1;patchIndex<=maxPatch;patchIndex++)
	{
		if(atomSize>=desiredSize)
		{
			maxAtom++;
			atomSize=0;
		}
		FEASSERT(patchIndex<=maxPatch);
		patchAtom[patchIndex]=maxAtom;

		FEASSERT(patchIndex<=primitiveCount);
		atomSize+=histogram[patchIndex];
#if FE_SAB_DEBUG
		feLog("patch %d atom %d\n",patchIndex,maxAtom);
#endif
	}

	std::vector<SurfaceRange::MultiSpan*> multiSpans(maxAtom+1,NULL);
	I32 spanAtom= -1;
	I32 spanStart= -1;
#if FE_SAB_DEBUG
	feLog("first %d from %d\n",spanAtom,spanStart);
#endif
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		FEASSERT(primitiveIndex<primitiveCount);
		const I32 patchIndex=primitivePatch[primitiveIndex];

		FEASSERT(patchIndex<0 || patchIndex<=maxPatch);
		const I32 atomIndex=patchIndex>=0? patchAtom[patchIndex]: -1;

		FEASSERT(atomIndex<=maxAtom);
		FEASSERT(spanAtom<=maxAtom);
		if(atomIndex!=spanAtom)
		{
			//* complete existing span
			SurfaceRange::MultiSpan* pMultiSpan=NULL;
			if(spanAtom>=0)
			{
				FEASSERT(spanAtom<=maxAtom);
				pMultiSpan=multiSpans[spanAtom];
				if(!pMultiSpan)
				{
					pMultiSpan=spanAtom? &spSurfaceRange->addAtomic():
							&spSurfaceRange->postAtomic();
					multiSpans[spanAtom]=pMultiSpan;
				}
			}
			if(spanStart>=0)
			{
#if FE_SAB_DEBUG
				feLog("complete %d span %d %d\n",
						spanAtom,spanStart,primitiveIndex-1);
#endif
				pMultiSpan->add(spanStart,primitiveIndex-1);
				spanStart= -1;
			}

			//* start another span
			spanAtom=atomIndex;
			if(atomIndex>=0)
			{
				spanStart=primitiveIndex;

#if FE_SAB_DEBUG
				feLog("start %d from %d\n",spanAtom,spanStart);
#endif
			}
		}
	}

	//* complete last span
	SurfaceRange::MultiSpan* pMultiSpan=NULL;
	if(spanStart>=0)
	{
		FEASSERT(spanAtom<=maxAtom);
		pMultiSpan=multiSpans[spanAtom];
		if(!pMultiSpan)
		{
			pMultiSpan=spanAtom? &spSurfaceRange->addAtomic():
					&spSurfaceRange->postAtomic();
			multiSpans[spanAtom]=pMultiSpan;
		}
		pMultiSpan->add(spanStart,primitiveCount-1);
	}

#if FE_SAB_DEBUG
	feLog("brief: %s\n",spSurfaceRange->brief().c_str());
	feLog("dump:\n%s\n",spSurfaceRange->dump().c_str());
#endif

	return spSurfaceRange;
}

} // namespace
